% Server code to connect to Browser
% SWI-Prolog (also handles variable nodes)
% non-portable: has_2nd_line (call)
% handles pair merge nb(ID,C1,C2)
% handles sets nb(ID,[C1,,,,Cn])
% handles non-overlapping terminals: overlap on/off in spaceBranching
% PROBLEM
%  backtrack?
%  TID
%
% WEBSOCKET DEBUGGING
%   assertz(ws_debug).
%   retractall(ws_debug).
%
% print_spellout(SO)	prints tree (with frontier) and spellout
%
% NEED TO LINK:
%   label_of(X,Label)
%   subconstituents(X,Cs)
%   subconstituent(X,C)
%   word(X,Word)
%   spellout(SO,Spellout)
%
%   newC(Value)		parse count value
%
%% (c) Sandiway Fong
%% University of Arizona (2014)
%% UPDATED: 7/31/2023

:- dynamic ws_debug/0.
% ws_debug.

:- use_module(library(http/websocket)).
:- use_module(library(http/thread_httpd)).
:- use_module(library(http/http_dispatch)).
% :- use_module(library(http/http_unix_daemon)).

start(Port) :- http_server(http_dispatch, [port(Port),workers(1)]).
stop(Port) :- http_stop_server(Port,[]).

% daemon :-
%     http_daemon([port(8010),
% 		 output('mm_output.txt'),
% 		 pidfile('mm_pid.txt')]).

% debug(http(_))
:- initialization(http_handler(root(ws),
			       http_upgrade_to_websocket(echo, []),
			       [spawn([])])).

:- dynamic ws/1.		% websocket
:- dynamic pending_step/0.

type(html).

echo(WebSocket) :-
    retractall(ws(_)),
    asserta(ws(WebSocket)),
    ws_receive(WebSocket, M),
    (M.opcode == close
    ->  format('Received close~n')
    ;   
    format('Received:~w~n',[M.data]),
    term_string(Term,M.data),
    call(Term)
    ).

print_spellout(X,S) :-
    output("Parse:"),
    newC(C),
    draw_tree2(X,C),
    output("Spell-out:"),
    spellout(X,S).

% stages: 
%  (1) extract frontier of parse
%  (2) apply head -> morpheme rules
%  (3) affix-hop
%  (4) apply head -> morpheme rules 2
%  (5) spell

% browser output
spellout(X,P) :-
    matrixC(X,Xp),
    frontier(Xp,_Occ,L,[]),
    with_output_to(string(Out),
%		   (format('<table class=sp><tr><td class=spo>'),
%		    print_list2(L),
%		    format('</td><td>(frontier)</td></tr>'))),
		   format('<table class=sp>')),
    output(Out),
    spellout_heads(L,M),	
    with_output_to(string(Out2),
		   (format('<tr><td class=spo>'),
		    print_list(M),
		    format('</td><td>(after morpheme realization)</td></tr>'))),
    output(Out2),
    affix_hop(M,N,[]),
    with_output_to(string(Out3),
		   (format('<tr><td class=spo>'),
		    print_list(N),
		    format('</td><td>(after affix-hop)</td></tr>'))),
    output(Out3),
    spellout2_heads(N,O),		
    with_output_to(string(Out4),
		   (format('<tr><td class=spo>'),
		    print_list(O),
		    format('</td><td>(after morpheme realization, stage 2)</td></tr>'))),
    output(Out4),
    spell(O,P),	
    with_output_to(string(Out5),
		   (format('<tr><td class=spo>'),
		    print_list(P),
		    format('</td><td>&nbsp;</td></tr>'))),
    output(Out5),
    with_output_to(string(Out6),format('</table>')),
    output(Out6).

set_deriv(X) :-
    atomics_to_string(["setDeriv(",X,");"],INS),
    process_in(INS).

start_encoder :-
    process_in("startEncoder();").

finish_encoder :-
    process_in("finishEncoder();").

output_link :-
    process_in("backToTop();").

output_settings(S) :-
    atomics_to_string(["writeSettings(\"",S,"\");"],INS),
    process_in(INS).

output('') :-
    !,				% red
    process_in("mtag(\" \",\"p\")").
output(S) :-
    (pending_step 		% for backtracking
     -> retract(pending_step),
	process_in("copy_append();",Step),
	(Step == 0 -> true ; step(_))			% depends on m10.pl
     ;  true),
    atomics_to_string(["mline(\"",S,"\");"],INS),
    process_in(INS).

printSuccess(S) :-
    atomics_to_string(["success(\"",S,"\");"],INS),
    process_in(INS).

output_error(S) :-
    atomics_to_string(["mline_error(\"",S,"\");"],INS),
    process_in(INS).

output_strong(S) :-
    atomics_to_string(["mline_strong(\"",S,"\");"],INS),
    process_in(INS).

output(S,Tag) :-
    atomics_to_string(["mtag(\"",S,"\",\"",Tag,"\");"],INS),
    process_in(INS).

backtrackstep(V) :-
%    assert(pending_step),	% new
    atomics_to_string(["backtrack(\"",V,"\");"],INS),
    process_in(INS).

% tree to be inserted into state table
% Step is nth step, used for id generation
draw_tree([],_) :- !. 		% red
draw_tree(T,Step) :-
    copy_term(T,T1),
    mark_copies(T1,T2),
    generateTreeID(Step,TID),
    atomics_to_string(["new_canvas(\"",TID,"\");"],INS),
    process_in(INS),
    leftMargin(N),
    spaceTree(T2,on,Tp),
    treeExtent(Tp,V,LE,RE,Depth),
    Width is RE - LE,
    resize(Depth,Width),
    V is -LE+N,
    placeTree(Tp),
    atomics_to_string(["save_canvas(\"",TID,"\");"],INS2),
    process_in(INS2),
    !.

generateTreeID(Step,TID) :-
    derivD(Deriv),
    atomics_to_string([Deriv,Step],TID).

% standalone tree
% presents the frontier too

draw_tree2([],_) :- !.		% red
draw_tree2(T,C) :-
    copy_term(T,T1),
    mark_copies(T1,T2),
    generateTreeID(C,TID),
    atomics_to_string(["new_standalone_canvas(\"",TID,"\");"],INS),
    process_in(INS),
    leftMargin(N),
    spaceTree(T2,off,Tp),
    treeExtent(Tp,V,LE,RE,Depth),
    Width is RE - LE,
    D is Depth + 2,		% allow room for spellout
    resize(D,Width),
    V is -LE+N,
    placeTree(Tp,1,Depth),
    atomics_to_string(["save_standalone_canvas(\"",TID,"\");"],INS2),
    process_in(INS2),
    !.

%    write(Tp), nl,  
%    format('V:~w LE:~w RE:~w Depth:~w~n',[V,LE,RE,Depth]),

process_ins :- ins(L), process_ins(L).

process_ins([]).
process_ins([I|Is]) :- process_in(I), process_ins(Is).

process_in(I) :-
    M = websocket{data:I,format:string,opcode:text},
    ws(WebSocket),
    (ws_debug -> format('Send:~w~n',[I]) ; true),
    ws_send(WebSocket, M),
    (ws_debug -> format('Sent~n',[]) ; true),
    ws_receive(WebSocket, Reply),
    (   Reply.opcode == close
    ->  format('Received close~n')
    ;   true).

% format('Send:~w~n',[I]),
%    format('Received:~w~n',[Reply.data])

process_in(I,Ans) :-
    M = websocket{data:I,format:string,opcode:text},
    ws(WebSocket),
    (ws_debug -> format('Send:~w~n',[I]) ; true),
    ws_send(WebSocket, M),
    (ws_debug -> format('Sent~n',[]) ; true),
    ws_receive(WebSocket, Reply),
    (   Reply.opcode == close, Ans = false
    ->  format('Received close~n')
    ;   (ws_debug -> format('Received:~w~n',[Reply.data]) ; true),
	(number_string(Ans,Reply.data) 
	 -> true 
	 ;  atom_string(Ans,Reply.data)
	)
    ).

%% primitives
%% phrase representation: n(Label,SubconstituentList)

subconstituents(Cs,X) :- cs(X,Cs).
subconstituent(Y,X) :- cs(X,Cs), member(Y,Cs).
label_of(X,L) :- label(X,L).
word(_X,_W) :- fail.

resize(Depth,Width) :-
    atomics_to_string(["resize(",Depth,",",Width,");"],INS),
    process_in(INS).

text(Text,Text,Width) :-
    string_concat("size(\"",Text,S),
    string_concat(S,"\");",INS),
    process_in(INS,Width).

moveNode(ID,X,N) :-
    atomics_to_string(["movenode(\"",ID,"\",",X,",",N,");"],INS),
    process_in(INS).

% with line extension (depth D) for terminal spellout
moveNode2(ID,X,N,D) :-
    atomics_to_string(["movenode2(\"",ID,"\",",X,",",N,",",D,");"],INS),
    process_in(INS).

nBranch(N,X,BF,Xs,_ID) :-
    atomics_to_string(["nbranch(",N,",",X,",",BF,","],S1),
    atomics_to_string(Xs,',',S2),
    string_concat(S1,S2,S3),
    string_concat(S3,");",INS),
    process_in(INS).

noBranch(N,X,_ID) :-
    atomics_to_string(["nobranch(",N,",",X,");"],INS),
    process_in(INS).

hBranch(N,Xs,_ID) :-
	atomics_to_string(["hzbranch(",N,","],S1),
	atomics_to_string(Xs,',',S2),
	string_concat(S1,S2,S3),
	string_concat(S3,");",INS),
	process_in(INS).

hBranch(N,X1,X2,_ID) :-
    atomics_to_string(["hbranch(",N,",",X1,",",X2,");"],INS),
    process_in(INS).

% display parameter (minimum branching in pixels)
:- dynamic branching/1.
branching(50).
original_branching(50).

% horizonal pixel separation between branches
:- dynamic horizSep/1.
horizSep(6).
original_horizSep(6).

% to be called by the tree display scale slider
setbranchsep(Mag) :-
    original_branching(BR),
    original_horizSep(HS),
    clause(branching(_),true,R1),
    erase(R1),
    B is BR*Mag,
    assert(branching(B)),
    clause(horizSep(_),true,R2),
    erase(R2),
    H is HS*Mag,
    assert(horizSep(H)).

leftMargin(10).

% spaceTree(Tree,on/off,INS)
% on/off: overlap
%% added set(L) nb(_,Cs)
spaceTree(VN,_,[v(2,E,E-1,E+1)]) :-
	variableNode(VN),
	!.
spaceTree(set(L),O,X) :-
	!,			% red
	X = [nb(_,Cs)|IDL],
	spaceTreeL(L,O,IDL),
	spaceBranching(IDL,O,b(_,_,_N,Cs)).
spaceTree(pm([T1,T2]),O,X) :-
    label_basetype(T1,H), 
    !,
    (headInitial(H) ->
	 X = [nb(_,C1,C2)|IDL],
	 spaceTreeL([T1,T2],O,IDL),
	 spaceBranching(IDL,O,b(_,_,2,[C1,C2]))
%	 format('head-initial C1:~w, C2:~w~n',[C1,C2])
     ;   X = [nb(_,C2,C1)|IDL],
	 spaceTreeL([T2,T1],O,IDL),
	 spaceBranching(IDL,O,b(_,_,2,[C2,C1]))).
%	 format('head-final C2:~w, C1:~w~n',[C2,C1])). % test
spaceTree(T,O,[v(ID,W,E,LE,RE)|Ra]) :- 
    pNode(T,ID,W,LW,RW),
    (has_2ndline(T,SL) ->
	 p2ndline(SL,ID2,W2,LW2,RW2),
	 Ra = [v(ID2,W2,E,E-LW2,E+RW2)|R]
     ;   Ra = R),
    (has_constituents(T,L) ->
	 spaceTreeL(L,O,IDL),
%	 (L = [pm(_)|_] -> gtrace ; true),
	 spaceBranching(IDL,O,LN),
	 R = [LN,IDL],
	 lineCenter(LN,E),
%	 (L = [pm(_)|_] -> format('LN:~w~nE(center):~w, W:~w~n',[LN,E,W]) ; true), % test
	 simplify(E-LW,LE),
	 simplify(E+RW,RE)
     ;   (has_constituent(T,T2) ->
	      spaceTree(T2,O,IDx),
	      unaryBranching(IDx,LN),
	      R = [LN,[IDx]],
	      lineCenter(LN,E),
	      simplify(E-LW,LE),
	      simplify(E+RW,RE)
	  ;   R = [],
	      LE = E-LW,
	      RE = E+RW)).

spaceTreeL([],_,[]).
spaceTreeL([T|Ts],O,[IDT|IDTs]) :-
	spaceTree(T,O,IDT),
	!,
	spaceTreeL(Ts,O,IDTs).

% Node X: ID
% total width: W, left width: LW, right width: RW
pNode(X,ID,W,LW,RW) :-
	ppNode(X,Label),
	text(Label,ID,W),
	LW is W // 2,
	RW is W - LW.

p2ndline(Label,ID,W,LW,RW) :-
	text(Label,ID,W),
	LW is W // 2,
	RW is W - LW.

unaryBranching(X,u(_,XC)) :- center(X,XC).

% line above branching is (center,left..right) in terms of variable X from
% rightmost of left of the pair
% b(ID,Mid,BF,[..C..])
% Mid = center point of branching
% BF = branching factor, C = Center position for each branch

spaceBranching(List,O,b(_,Mid,BF,Cs)) :-
	horizSep(N),
	length(List,BF),
	(O == on ->
	     spaceBranchingL(List,N,Cs)
	 ;  spaceBranchingLnover(List,N,Cs)),
	first(Cs,XC),
	last(Cs,YC),
	midPoint(XC,YC,Mid).

spaceBranchingLnover([_],_,[]).
spaceBranchingLnover([L,R|Listp],N,[XC,YC|Cs]) :-
    rightExtent(L,XC,_),
    leftExtent(R,YC,_),
    rightmost(L,LN),
    rightExtent([LN],_,Lr),
    leftmost(R,RN),
    leftExtent([RN],_,Rr),
    eval(Rr,Lr+N,F),
    minBranching(XC,YC,F),
    !,
    spaceBranchingLnover([R|Listp],N,Cs).

% need to maximize F through levels
spaceBranchingL([L,R|Listp],N,[XC,YC|Cs]) :-
	rightExtent(L,XC,XE),
	leftExtent(R,YC,YE),
	eval(YE,XE+N,F),
	rights(L,[_|REs]),
	lefts(R,[_|LEs]),
	recurseRL(REs,LEs,F,N,Fp), 
	minBranching(XC,YC,Fp),
	!,
	spaceBranchingL2([R|Listp],N,REs,Cs).

spaceBranchingL2([_],_,_,[]).
spaceBranchingL2([M,R|Listp],N,REs,[ZC|Cs]) :-
	rightExtent(M,MC,ME),
	leftExtent(R,ZC,ZE),
	eval(ZE,ME+N,F2),
	rights(M,[_|MEs]),
	mergeRights(MEs,REs,MREs),
	lefts(R,[_|ZEs]),
	recurseRL(MREs,ZEs,F2,N,F2p), 
	minBranching(MC,ZC,F2p),
	!,
	spaceBranchingL2([R|Listp],N,MREs,Cs).

% Primary is first, secondary is 2nd arg
% e.g. primary=[1,2,3] secondary=[a,b,c,d,e], merged=[1,2,3,d,e]
% e.g. primary=[1,2,3] secondary=[a,b], merged=[1,2,3]

mergeRights([],Rs,Rs).
mergeRights([H|T],Rs,[H|T2]) :-
	cdrNF(Rs,Rsp),
	mergeRights(T,Rsp,T2).

cdrNF([],[]).
cdrNF([_|T],T).

minBranching(XC,YC,Fp) :-
	branching(B),
	copy_term(XC,XC1),
	copy_term(YC,YC1),
	exprVar(YC1,Y1),
	exprVar(XC1,X1),
	Y1 = X1+Fp,
	simplify(XC1,0,M),
	simplify(YC1,0,N),
	Bp is N - M,
	(B > Bp -> D is B-Bp+Fp ; D = Fp),
	exprVar(YC,Y),
	exprVar(XC,X),
	Y = X + D.

midPoint(XC,YC,V+M) :-
	simplify(XC,0,L,V),
	simplify(YC,0,R),
	M is (L+R) // 2.
	
recurseRL(Ls,Rs,F,N,Fp) :-
	(Ls=[];Rs=[])
	->  Fp = F
	;   Ls = [XE|Lsp],
	    Rs = [YE|Rsp],
	    eval(YE,XE+N,F2),
	    (F2 > F
	    ->  recurseRL(Lsp,Rsp,F2,N,Fp)
	    ;   recurseRL(Lsp,Rsp,F,N,Fp)).

% (rightmost) XE = YE (leftmost)
% compute XE-YE for variables set to zero
% establishes Y = X + F relation between XE and YE variables
eval(YE,XE,F) :-
	copy_term(XE,XEp),
	copy_term(YE,YEp),
	exprVar(XEp,0),
	exprVar(YEp,0),
	F is XEp-YEp.
%	write(YE), write(' '), write(XE), write(' '), write(F), nl.

exprVar(E,V) :- term_variables(E,[V]).

% for nodes: v(ID,Width,Var,Var-Left,Var+Right)
% v/4 for variable nodes
% for branching: b(ID,Var,[Var-Left..Var+Right])
% for nonbranching: u(ID,Var)
% for nb/2,3, we use the left edge always, why not do as n-ary b?
leftExtent([nb(_,[X|_])|_],X,X).
leftExtent([nb(_,X,_)|_],X,X).
leftExtent([[v(_,_,X,L,_)|_]|_],X,L).
leftExtent([v(_,_,X,L,_)|_],X,L).
leftExtent([v(_,X,L,_)|_],X,L).
leftExtent([b(_,X,_,[L|_])|_],X,L).
leftExtent([u(_,X)|_],X,X).
rightExtent([nb(_,[X|_])|_],X,X).
rightExtent([nb(_,X,_)|_],X,X).
rightExtent([[v(_,_,X,_,R)|_]|_],X,R).
rightExtent([v(_,_,X,_,R)|_],X,R).
rightExtent([v(_,X,_,R)|_],X,R).
rightExtent([b(_,X,_,Cs)|_],X,R) :- last(Cs,R).
rightExtent([u(_,X)|_],X,X).
center([v(_,_,X,_,_)|_],X).
center([v(_,X,_,_)|_],X).
center([b(_,X,_,_)|_],X).
center([u(_,X)|_],X).
center([nb(_,X,_)|_],X).
center([nb(_,[X|_])|_],X).

% going down the left and right branches, resp.
% Cases
% [V,FB,[T1,T2,T3,T4,T5]] [FB,[T1,T2,T3,T4,T5]]
% [V,QB,[T1,T2,T3,T4]] 	[QB,[T1,T2,T3,T4]]
% [V,TB,[T1,T2,T3]] 	[TB,[T1,T2,T3]]
% [V,B,[T1,T2]] 	[B,[T1,T2]]
% [V,U,T]		[U,T]
% [V,V]			[V]
% [FB,[T1,T2,T3,T4,T5]]	T1 for leftmost, T5 for rightmost
% [QB,[T1,T2,T3,T4]]	T1 for leftmost, T4 for rightmost
% [TB,[T1,T2,T3]]	T1 for leftmost, T3 for rightmost
% [B,[T1,T2]]		T1 for leftmost, T2 for rightmost
% [U,T]			T

childLeft([NB,[X|_]],X) :- nb(NB), !.	    % green
childLeft([B,[T|_]],T) :- bbOrHigher(B), !. % green
childLeft([L],C) :- child(L,C), !. % green
childLeft(L,C) :- child(L,C), !. % green
childLeft(_,[]).

childRight([NB,List],R) :- nb(NB), !, last(List,R). % green
childRight([B,List],T) :- bbOrHigher(B), !, last(List,T). % green
childRight([L],C) :- child(L,C), !. % green
childRight(L,C) :- child(L,C), !. % green
childRight(_,[]).

lefts(N,[LE|LEs]) :-
	leftExtent(N,_,LE),
	((N=[V,BR,L], node(V), bbOrHigher(BR)) ->
	    leftExtent([BR,L],_,LE2),
	    LEs = [LE2|LEsp],
	    ascendingLefts(L,LEsp)
	;   ((N=[V,U,[T]], node(V), ub(U)) ->
		leftExtent([U|_],_,LE2),
		LEs = [LE2|LEsp],
		lefts(T,LEsp)
	    ;   ((N=[V1,V2], node(V1)) ->
		    leftExtent([V2],_,LE2), LEs=[LE2]
		;    (N=[[_V1,V2]] ->
			 leftExtent([V2],_,LE2), LEs=[LE2]
		      ;   ((N = [NB,L|_], nb(NB)) ->
			       lefts(L,LEs)
			   ; LEs= []))))).

rights(N,[LE|LEs]) :-
	rightExtent(N,_,LE),
	((N=[V,BR,L], node(V), bbOrHigher(BR)) ->
	    rightExtent([BR,L],_,LE2),
	    LEs = [LE2|LEsp],
	    ascendingRights(L,LEsp)
	;   ((N=[V,U,[T]], node(V), ub(U)) ->
		rightExtent([U|_],_,LE2),
		LEs = [LE2|LEsp],
		rights(T,LEsp)
	    ;   ((N=[V1,V2], node(V1)) ->
		    rightExtent([V2],_,LE2), LEs=[LE2]
		;    (N=[[_V1,V2]] ->
			 rightExtent([V2],_,LE2), LEs=[LE2]
		     ;   ((N = [NB|Xs], nb(NB)) ->
			  last(Xs,L),
			  rights(L,LEs)
			 ; LEs= []))))).

leftmost(N,Lm) :-
    ((N=[V,BR,L], node(V), bbOrHigher(BR)) ->
	 first(L,F),
	 leftmost(F,Lm)
     ;   ((N=[V,U,[T]], node(V), ub(U)) ->
	      leftmost(T,Lm)
	  ;   ((N=[V1,V2], node(V1)) ->
		   leftmost(V2,Lm)
	       ;  (N=[[_V1,V2]] ->
		       leftmost(V2,Lm)
		   ;   ((N = [NB,L|_], nb(NB)) ->
			    leftmost(L,Lm)
			   ; Lm = N))))).

rightmost(N,Lm) :-
    ((N=[V,BR,L], node(V), bbOrHigher(BR)) ->
	 last(L,F),
	 rightmost(F,Lm)
     ;   ((N=[V,U,[T]], node(V), ub(U)) ->
	      rightmost(T,Lm)
	  ;   ((N=[V1,V2], node(V1)) ->
		   rightmost(V2,Lm)
	       ;  (N=[[_V1,V2]] ->
		       rightmost(V2,Lm)
		   ;   ((N = [NB|Xs], nb(NB)) ->
			last(Xs,L),
			rightmost(L,Lm)
		       ; Lm = N))))).
% list of nodes
ascendingLefts([],[]) :- !.
ascendingLefts(List,LEs) :-
	depths(List,Ds),
	filterAscendingR(List,Ds,0,Listp),
	Listp =[Leftmost|_],
	leftExtent(Leftmost,_,LE),
	LEs = [LE|LEsp],
	descendOnceL(Listp,Listpp),
	ascendingLefts(Listpp,LEsp).

ascendingRights([],[]) :- !.
ascendingRights(List,LEs) :-
	depths(List,Ds),
	filterAscendingL(List,Ds,0,Listp),
	last(Listp,Rightmost),
	rightExtent(Rightmost,_,LE),
	LEs = [LE|LEsp],
	descendOnceL(Listp,Listpp),
	ascendingRights(Listpp,LEsp).

% keep K1.. K2 if K2 > K1 depth-wise, i.e. find ascenders to the right
filterAscendingR([],[],_,[]).
filterAscendingR([H|T],[D|Ds],N,L) :-
	(D > N ->
	    L = [H|Lp],
	    filterAscendingR(T,Ds,D,Lp)
	;   filterAscendingR(T,Ds,N,L)).

% keep K2.. K1 if K2 > K1 depth-wise, i.e. find ascenders to the left
filterAscendingL(L,Ds,0,Lp) :-
	reverse(L,Ds,[],[],Lv,Dsv),
	filterAscendingR(Lv,Dsv,0,Lvp),
	reverse(Lvp,[],Lp).

% list of nodes, move down one
descendOnceL([],[]).
descendOnceL([H|T],L) :-
	((H = [B,L1], bbOrHigher(B)) ->
	    append(L1,L2,L),
	    descendOnceL(T,L2)
	 ;   ((H = [NB|Ts], nb(NB)) ->
		  append(Ts,L2,L),
		  descendOnceL(T,L2)
	      ;   children(H,L1),
		  (L1 == [] ->
		       descendOnceL(T,L)
		   ;   L = [L1|Ls],
		       descendOnceL(T,Ls)))).

% move down one (return single node)
children([V,B|R],[B|R]) :- node(V), bbOrHigher(B).
children([V,U|R],[U|R]) :- node(V), ub(U).
children([U,[N]],N) :- ub(U).
children([V1,V2],[V2]) :- node(V1), node(V2).
children([V],[]) :- node(V).

depths([],[]).
depths([H|T],[D|Ds]) :- depth(H,D), depths(T,Ds).

depth([V,B,TL],D) :- node(V), bbOrHigher(B), maxDepth(TL,0,D1), D is D1+2.
depth([V,U,T],D) :- node(V), ub(U), depth(T,D1), D is D1+2.
depth([V1,_],2) :- node(V1).
depth([B,TL],D) :- bbOrHigher(B), maxDepth(TL,0,D1), D is D1+1.
depth([U,T],D) :- ub(U), depth(T,D1), D is D1+1.
depth([NB|Ts],D) :- nb(NB), maxDepth(Ts,0,D1), D is D1+1.
depth([[V1,_]],2) :- node(V1).
depth([V1],1) :- node(V1), !. 	% red
depth([T],D) :- depth(T,D).

maxDepth([],D,D).
maxDepth([C|Cs],D1,D) :-
	depth(C,D2),
	(D1 > D2 ->
	    maxDepth(Cs,D1,D)
	;   maxDepth(Cs,D2,D)).
	
child([V,B,TL],[B,TL]) :- node(V), bbOrHigher(B).
child([V,U,T],[U,T]) :- node(V), ub(U).
child([V1,V2],[V2]) :- node(V1).
child([U,T],T) :- ub(U).
child([NB|Ts],Tsp) :- nb(NB), (Ts = [T] -> Tsp = T ; Tsp = Ts).

node(V) :- functor(V,v,5).
node(V) :- functor(V,v,4).	% variable node (see 1st clause of spaceTree/2)
ub(U) :- functor(U,u,2).
bb(B) :- functor(B,b,4).
nb(NB) :- functor(NB,nb,3).
nb(NB) :- functor(NB,nb,2).	% for sets

% binary branching or higher branching
bbOrHigher(B) :- functor(B,b,4).

% for ternary branching: (Center,Left,Middle,Right)
% for binary branching: (Center,Left,Right)
% no branching: (Center)
% for vertex: (Center)
% for nb, left is Center
lineCenter(b(_,C,_,_),C).
lineCenter(u(_,C),C).
lineCenter(nb(_,C,_),C).
lineCenter(nb(_,[C|_]),C).
lineCenter(v(_,_,C,_,_),C).

% computes V-LE and V+RE for the whole tree
% feed it the output of spaceTree/2
% counting at 3: one blank at top, one at bottom, one is first node

% 5 parameter version
treeExtent([nb(_,_)|List],Vr1,LE1,RE2,D) :- % not sure, call treeExtentsL instead?
	append([First|Others],[Last],List),
	treeExtent(First,Vr1,LE1,_,D1),
	treeDepthsL(Others,D1,D2), % based on nb/3
	treeExtent(Last,_,_,RE2,D3),
	max(D3,D2,D).
treeExtent([nb(_,_,_),[V1|L1],[V2|L2]],Vr1,LE1,RE2,D) :-
    treeExtent([V1|L1],Vr1,LE1,_,D1),
    treeExtent([V2|L2],_,_,RE2,D2),
    max(D1,D2,D).
treeExtent([v(_,_,_,L,R)|T],V,LEp,REp,D) :-
	simplify(L,0,LE,V),
	simplify(R,0,RE,_),
	treeExtent(T,LE,LEp,RE,REp,3,D). % calls 7 parameter version

treeDepthsL([],D,D).
treeDepthsL([H|T],D1,D3) :-
	treeExtent(H,_,_,_,D2),
	(D2 > D1 -> 
	 treeDepthsL(T,D2,D3)
	;
	 treeDepthsL(T,D1,D3)).

% 7 parameter version
treeExtent([],LE,LE,RE,RE,N,N) :-
	!.
treeExtent([v(_,_,_,_)],LE,LE,RE,RE,N,M) :-
	!,
	M is N+1.
treeExtent([v(_,_,_,L,R)|T],LE,LEp,RE,REp,N,M) :-
	!,
	simplify(L,0,LE1),
	simplify(R,0,RE1),
	updateLE(LE,LE1,LE2),
	updateRE(RE,RE1,RE2),
	M1 is N+1,
	treeExtent(T,LE2,LEp,RE2,REp,M1,M).
treeExtent([b(_,_,_,Cs),T],LE,LEp,RE,REp,N,M) :-
	!,
	first(Cs,L),
	last(Cs,R),
	simplify(L,0,LE1),
	simplify(R,0,RE1),
	updateLE(LE,LE1,LE2),
	updateRE(RE,RE1,RE2),
	M1 is N+1,
	treeExtentL(T,LE2,LEp,RE2,REp,M1,M).
treeExtent([nb(_,_)|List],LE,LEp,RE,REp,N,M) :-
	!,
	M1 is N+1,
	treeExtentL(List,LE,LEp,RE,REp,M1,M).
treeExtent([nb(_,_,_),T1,T2],LE,LEp,RE,REp,N,M) :-
	!,
	M1 is N+1,
	treeExtentL([T1,T2],LE,LEp,RE,REp,M1,M).
treeExtent([u(_,_),T],LE,LEp,RE,REp,N,M) :-
	!,
	M1 is N+1,
	treeExtentL(T,LE,LEp,RE,REp,M1,M).
	
updateLE(LE,LE1,LEp) :- LE1 < LE -> LEp = LE1 ; LEp = LE.
updateRE(RE,RE1,REp) :- RE1 > RE -> REp = RE1 ; REp = RE.

treeExtentL([],LE,LE,RE,RE,N,N) :- !.
treeExtentL([H|T],LE,LEp,RE,REp,N,Np) :-
	treeExtent(H,LE,LE1,RE,RE1,N,M1),
	treeExtentL(T,LE1,LEp,RE1,REp,N,M2),
	gt(M1,M2,Np).

gt(M1,M2,N) :- M1 > M2 -> N=M1 ; N=M2.

% assumes ((V+/-K)+/-K) etc. i.e. variable is most deeply embedded in left
% adds up all the +/-K and tidys it up as V+/-K
simplify(E,SE) :- simplify(E,0,N,V), sum(V,N,SE).

simplify(V,N,N,V) :- var(V), !.
simplify((X+Y),N,M,V) :- nonvar(Y), NY is N+Y, simplify(X,NY,M,V).
simplify((X-Y),N,M,V) :- nonvar(Y), NY is N-Y, simplify(X,NY,M,V).

sum(V,N,SE) :- N < 0 -> MN is -N, SE = V-MN ; SE = V+N.

simplify(V,N,N) :- var(V), !.
simplify((X+Y),N,M) :- nonvar(Y), NY is N+Y, simplify(X,NY,M).
simplify((X-Y),N,M) :- nonvar(Y), NY is N-Y, simplify(X,NY,M).

%%% Move nodes to their correct positions on display

% movenode ID X N
% branch N Center Left Right
% nbranch N Center BF [Left..Right]
placeTree(T) :-	placeTree(T,1).

placeTreeL([],_).
placeTreeL([H|T],N) :- placeTree(H,N), placeTreeL(T,N).

placeTree([],_).
placeTree([v(_,_,_,_)],_).	% variable node
placeTree([v(ID,_,C,_,_)|T],N) :-
    !,			% red
    X is C,
    moveNode(ID,X,N),
    M is N+1,
    placeTree(T,M).
placeTree([b(ID,C,BF,Cs)|T],N) :-
    !,			% red
    X is C,
    isL(Cs,Xs),
    nBranch(N,X,BF,Xs,ID),
    M is N+1,
    placeTree(T,M).
placeTree([u(ID,C)|T],N) :-
    !,			% red
    X is C,
    noBranch(N,X,ID),
    M is N+1,
    placeTree(T,M).
placeTree([nb(ID,Cs)|T],N) :-
    !,			% red
    isL(Cs,Xs),
    hBranch(N,Xs,ID),
    M is N+1,
    placeTree(T,M).
placeTree([nb(ID,C1,C2)|T],N) :-
    !,			% red
    X1 is C1,
    X2 is C2,
    hBranch(N,X1,X2,ID),
    M is N+1,
    placeTree(T,M).
placeTree(List,N) :- placeTreeL(List,N).

% nodes x level x depth
% used for node placement plus frontier extension
placeTree([],_,_).
placeTree([v(_,_,_,_)],_,_).	% variable node
placeTree([v(ID,_,C,_,_)|T],N,D) :-
    !,			% red
    X is C,
    moveNode(ID,X,N),
    (T == []
     -> moveNode2(ID,X,N,D)
     ;  true),
    M is N+1,
    placeTree(T,M,D).
placeTree([b(ID,C,BF,Cs)|T],N,D) :-
    !,			% red
    X is C,
    isL(Cs,Xs),
    nBranch(N,X,BF,Xs,ID),
    M is N+1,
    placeTree(T,M,D).
placeTree([u(ID,C)|T],N,D) :-
    !,			% red
    X is C,
    noBranch(N,X,ID),
    M is N+1,
    placeTree(T,M,D).
placeTree([nb(ID,Cs)|T],N,D) :-
	!,			% red
	isL(Cs,Xs),
	hBranch(N,Xs,ID),
	M is N+1,
	placeTree(T,M,D).
placeTree([nb(ID,C1,C2)|T],N,D) :-
    X1 is C1,
    X2 is C2,
    hBranch(N,X1,X2,ID),
    M is N+1,
    placeTree(T,M,D).
placeTree(List,N,D) :- placeTreeL(List,N,D).

placeTreeL([],_,_).
placeTreeL([H|T],N,D) :- placeTree(H,N,D), placeTreeL(T,N,D).

% conversion into a number by evaluation of expression X
isL([],[]).
isL([C|Cs],[X|Xs]) :- X is C, isL(Cs,Xs).

%%% Multiword items

mWord(List,MWord) :- multiWord(List,MWord).
multiWord(List,MWord) :- atomic_list_concat(List,' ',MWord).

%%% Phrase Structure for tree layout

ppNode(X,LB) :- 
    label_of(X,LB1), 
    (var(LB1) -> LB = ' ' ; LB = LB1).

variableNode(V) :- var(V).

has_constituent(X,Y) :- subconstituents([Y],X).
has_constituents(X,L) :- subconstituents(L,X), L = [_,_|_].

% 2nd line: non-portable
has_2ndline(X,SL) :- word(X,SL).

%%% List stuff

first([X|_],X).

reverse([],[],L1,L2,L1,L2).
reverse([H1|T1],[H2|T2],L1,L2,R1,R2) :- reverse(T1,T2,[H1|L1],[H2|L2],R1,R2).

reverse([],L,L).
reverse([H|T],L,R) :- reverse(T,[H|L],R).

% Misc

max(X,Y,Z) :- X > Y -> Z = X ; Z = Y.
